use std::{
	env, fs,
	sync::Mutex,
	fs::{
		ReadDir
	},
	path::Path,
};

use actix_web::{
	get, web,
	App, HttpResponse, HttpServer, Responder,
};

mod lib;
use lib::{
	*,
	Archiveable,
	boards::{
		Board,
		BoardFactory
	},
	users::UserManager,
	messages::MessageFactory
};

mod api;
use api::types::AppState;

mod html;
use html::{
	accounts::*,
	boards::*
};

/*
struct PageBuilder {
	body: &'static str,
	style: Option<String>,
	table_labels: Option<String>
}

impl PageBuilder {
	fn new(text: &'static str) -> PageBuilder {
		PageBuilder {
			body: text,
			style: None,
			table_labels: None
		}
	}

	fn build<S>(self, style: S, table_labels: S) -> String where S: AsRef<str> {
		String::from(self.body)
			.replacen("{}", style.as_ref(), 1)
			.replacen("{}", table_labels.as_ref(), 1)
	}
}
*/

#[get("/")]
async fn index() -> impl Responder {
	HttpResponse::Ok().body(r#"
<html><head>
<link rel="stylesheet" type="text/css" href="https://orbitalfox.us/Music/index.css">
</head>
<body>
<h1>Boarders v1.0.0</h1>
<p>
<h2>What is it?</h2>
Boarders is an easily deployable message board written in Rust.
It is extremely efficient on both the frontend and backend.
On the frontent all pages are static and (at this time) have no JS, only CSS/HTML.
The backend uses the 5th fastest web framework, Actix, according to <a href="https://www.techempower.com/benchmarks/">techempower.com</a>.
Which ensures that response times will always be snappy.
</p>
<p>
<hr>
<h2>Rules</h2>
<ol>
<li>No NSFW<br>
At least until NSFW channels are implemented</li>
<li>Don't spam boards or user accounts<br>
It's just anoying, and if it gets too bad I may have to hard reset everything<br>
Don't ruin the fun for everyone</li>
<li>Have Fun!<br>Pretty much anything goes, but if people don't like you, that's your fault</li>
</ol>
</p>
<h3 style="text-align: center;">------<a href="/boards/">[Get Started]</a>------</h3>
</body>
</html>
"#)
}

#[get("/debug/")]
async fn debug(
	data: web::Data<AppState>
) -> impl Responder {
	HttpResponse::Ok().body(serde_json::to_string_pretty(data.as_ref()).unwrap())
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
	println!("Initializing...");
	let data_dir = Path::new("data/");
	let mut pwd = env::current_dir().unwrap();

	let board_files: ReadDir = match fs::read_dir({
		pwd.push(data_dir);
		pwd.push("boards/");
		&pwd
	}) {
		Ok(a) => a,
		Err(_e) => {
			fs::create_dir_all(&pwd).unwrap();
			fs::read_dir(pwd).unwrap()
		}
	};

	let mut boards: Vec<Board> = Vec::new();

	for file in board_files.map(|x| x.unwrap()) {
		let mut path = file.path().to_path_buf();
		//let meta = fs::metadata(&path).unwrap();
		let ftype = file.file_type().unwrap();
		if ftype.is_dir() {
			path.push("board.json");
			println!("loading board: {:?}", path);
			boards.push(
				Board::open(&path).unwrap_or_else(|e| {
					panic!("Could not open board {:?}\n{}", path, e)
				})
			);
		}
	}

	boards.sort_by(|a,b| a.id.cmp(&b.id));

	let mut board_data = std::path::PathBuf::new();
	board_data.push(&data_dir);
	let ff = FactoryFactory::new(data_dir);
	let (mut bf, mf, um) = ff.build();
	board_data.push("boards/");
	bf.set_data_dir(board_data);

	let web_data = web::Data::new(AppState {
		boards: Mutex::new(boards.clone()),
		board_factory: Mutex::new(BoardFactory::load_state({
			println!("Loading board factory file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("boards/factory.json");
			dd
		}).unwrap_or(bf)),
		msg_factory: Mutex::new(MessageFactory::load_state({
			println!("Loading message factory file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("message_factory.json");
			dd
		}).unwrap_or(mf)),
		user_manager: Mutex::new(UserManager::load_state({
			println!("Loading user manager file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("users.json");
			dd
		}).unwrap_or(um))
	});
	let wb = web_data.clone();

	println!("Finished initialization");
	println!("Starting server...");

	let server = HttpServer::new(move || {
		App::new()
			.app_data(web_data.clone())
			.service(index)
			.service(debug)
			.service(list_boards)
			.service(login)
			.service(auth_html)
			.service(sign_up)
			.service(sign_up_result)
			.service(new_board)
			.service(new_board_result)
			.service(get_board)
			.service(board_post)
			.service(
				web::scope("/api")
					.service(api::boards::new)
					.service(api::boards::list)
					.service(api::messages::new)
					.service(api::users::new)
					.service(api::users::list)
					.service(api::users::get)
					.service(api::users::auth))
			.service(actix_files::Files::new("/static/", "./static/").show_files_listing())
	}).bind("127.0.0.1:8080")?
		.run();
	println!("Started server");
	println!("{}", include_str!("res/banner"));
	let res = server.await;
	println!("\nExiting...");
	/*
	 * Note to self: handle all errors as much as possible
	 * Use unwrap() as little as possible
	 * If anything goes wrong while saving data data WILL get lost
	 */
	println!("Saving application state...");
	for b in wb.boards.lock().unwrap().iter() {
		fs::create_dir_all(&b.path).unwrap_or_else(
			|x| { println!("Could not create directory {:?}: {}", b.path, x) }
		);
		b.save_state("board.json").unwrap_or_else(
			|x| { println!("Could not save board {:?}: {}", b.title, x) }
		);
	}

	wb.board_factory
		.lock()
		.unwrap()
		.save_state("factory.json")
		.unwrap_or_else(
			|x| { println!("Could not save user factory state: {}", x) }
		);

	wb.msg_factory
		.lock()
		.unwrap()
		.save_state("message_factory.json")
		.unwrap_or_else(
			|x| { println!("Could not save message factory state: {}", x) }
		);

	wb.user_manager
		.lock()
		.unwrap()
		.save_state("users.json")
		.unwrap_or_else(
			|x| { println!("Could not save user data: {}", x) }
		);
	println!("Done!");
	res
}
